#!/usr/bin/env python
import math, sys, os
import subprocess
import socket
import logging
import signal
import time
import RPi.GPIO as GPIO
import alsaaudio
import linphone

# RaRoPhoPl - the internet sip/voip RAspberry pi ROtary PHOne with mp3 PLayer/internet/web radio (python, linphone, ...)
#
# This art piece is provided for studying / research purposes only
# based on: http://www.instructables.com/id/Fairytale-Phone/
# and: https://wiki.linphone.org/wiki/index.php/Raspberrypi:start
# and a Raspbian from June 2015 with Python2... 
#
# Take the chance to read this as a special gift *without* any warranty / guarantee or liability. 
# ( If You want to give me a gift too: 
#   https://medienwunder.wordpress.com/spenden-donate/
#   Thank You for Your support! ;-)
#  )
#
# Please keep in mind that there is NO service, NO support, NO further information, NO warranty, NO guarantee, NO liability, ... 
# - and everything is on Your own risk! So be aware: if You start this as a program there could be some harm to something were nobody is accountably but You!
#
# Rotary:
# on hook action: mpg123: 1-9 sounds/stations, 0 switch volume (five times = system halt), # button could be system halt too
# off hook action: phone: 0-9 to dial (wait after last number) with no 0 = '**' else '*31*', > 0 can add local area code  
# see
# https://medienwunder.wordpress.com/2015/06/14/rarophopl-allip-sip-voip-internet-radio-und-waehlscheiben-telefon-mit-impulswahl-raspberry-pi-fritzbox-python-linphone/ 
# (or if it's gone - search for it under: https://einoe.de/x/mb )
# for more instructions...
#
# Fritz!Box: Check your manual for the right domain settings (here: fritz.box but e.g. on some models: fritz.fonwlan.box)
# Use find & replace if needed...
#
# I would suggest to START with "main" at the very ending of the file (there are the (account-)settings, too).
#
# I do NOT allow You to sell this or new pieces based on my work. 
# Enjoy and share your improvements with the community, please. Thank You!
#
# DEUTSCH: KEIN Service, KEINE Updates, KEINE weiteren Informationen, KEINE Garantie für irgendwas! 
#          KEIN Verkauf! Nur als Anregung, NICHT als Anwendung gedacht!
#
# Good luck! :-) 
# https://einoe.de/x/mb 
#

GPIO.cleanup()
GPIO.setwarnings(False)
GPIO.setmode(GPIO.BCM)  
GPIO.setup(18, GPIO.IN, pull_up_down=GPIO.PUD_UP)	# Rotary use - PUD_UP to Ground (PUD_DOWN to 3,3V)
GPIO.setup(23, GPIO.IN, pull_up_down=GPIO.PUD_UP)	# Rotary count
GPIO.setup(24, GPIO.IN, pull_up_down=GPIO.PUD_UP)	# Extra button
GPIO.setup(25, GPIO.IN, pull_up_down=GPIO.PUD_UP)	# Hook

c = 0
offswitch = 0
timestart = 0
telstat = 0 # 0 = wait for action, 1 = start call, 2 = in call, 3 = hang up, 4 = get call, 5 = decline
music = ''  # startmusic
# play startmusic
# player = subprocess.Popen(["mpg123", "-q", music], stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

class RotaryPlayerPhone:
	def __init__(self, username='', password='', whitelist=[], camera='', snd_capture='', snd_playback=''):
		self.quit = False
		self.whitelist = whitelist
		callbacks = {
			'call_state_changed': self.call_state_changed,
		}

		# Configure the linphone core
		logging.basicConfig(level=logging.INFO)
		signal.signal(signal.SIGINT, self.signal_handler)
		linphone.set_log_handler(self.log_handler)
		self.core = linphone.Core.new(callbacks, None, None)
		self.core.max_calls = 1
		self.core.echo_cancellation_enabled = False
		self.core.video_capture_enabled = False
		self.core.video_display_enabled = False
		self.core.inc_timeout = 60  # If an incoming call isn't answered for this timeout period, it is automatically declined.
		self.core.ring_during_incoming_early_media = True  #[bool] Enable or disable the ring play during an incoming early media call.
		# self.core.stun_server = 'fritz.box'
		# self.core.firewall_policy = linphone.FirewallPolicy.PolicyUseIce
		if len(camera):
			self.core.video_device = camera
		if len(snd_capture):
			self.core.capture_device = snd_capture
			self.core.ringer_device = snd_capture	#[string] Sets the sound device used for ringing (and busy signal? So changed it below for my construction...)
		if len(snd_playback):
			self.core.playback_device = snd_playback

		# Only enable PCMU and PCMA audio codecs
		for codec in self.core.audio_codecs:
			if codec.mime_type == "PCMA" or codec.mime_type == "PCMU":
				self.core.enable_payload_type(codec, True)
			else:
				self.core.enable_payload_type(codec, False)

		# Only enable VP8 video codec
		for codec in self.core.video_codecs:
			if codec.mime_type == "VP8":
				self.core.enable_payload_type(codec, True)
			else:
				self.core.enable_payload_type(codec, False)

		# Only change/enable if something else than the default is needed
		self.core.ring = '/usr/local/lib/python2.7/dist-packages/linphone/unittests/sounds/oldphone.wav'	#[string] path to a wav file used for ringing.
		#self.core.ring_level = 100	#[int] Set sound ring level in 0-100 scale
		#self.core.ringback	= ''	#[string] Sets the path to a wav file used for ringing back (means the ring that is heard when it's ringing at the remote party. The file must be a wav 16bit linear.)
		#self.core.remote_ringback_tone = '/usr/local/lib/python2.7/dist-packages/linphone/unittests/sounds/ringback.wav'  # [string] Specify a ring back tone to be played to far end during incoming calls.

		self.configure_sip_account(username, password)
		

	def signal_handler(self, signal, frame):
		self.core.terminate_all_calls()
		# write log-file
		myfile = file('playerphone.log', 'a')   # start log: log-file-handle
		mystring = 'RaRoPhoPl ended %s: %s \n' % (telstat, time.asctime( time.localtime(time.time()) ) )
		myfile.write(mystring)
		myfile.close()	# end log
		GPIO.cleanup()
		os.system("amixer -c 0 sset Headphone 98%,98%")
		self.quit = True
		

	def log_handler(self, level, msg):
		method = getattr(logging, level)
		method(msg)
		

	def call_state_changed(self, core, call, state, message):
		global telstat
		global timestart
		timecount = 0
		timecnt = 0
		timest = 0
		timestart = time.time()
		if state == linphone.CallState.IncomingReceived:
			core.accept_early_media(call)
			if call.remote_address.as_string_uri_only() in self.whitelist:
				telstat = 3
				while (telstat < 4):
					timecount = time.time() - timestart
					# From hook at incoming call set telstat = 4, decline 5 
					try:
						# if GPIO.event_detected(25):   # short version: hook switch
						#   telstat = 4
						if (GPIO.input(25) != 0):   # long version: still on hook?
							if (timest == 0):
								timest = timecount
							timecnt = timecount - timest
							if (timecnt > 1):
								timest = 0
								timecnt = 0
								telstat = 3  
						else:   # Off hook
							if (timest == 0):
								timest = timecount
							timecnt = timecount - timest
							if (timecnt > 1):
								timest = 0
								timecnt = 0
								telstat = 4         # End of long version
								print ("**** Incoming: Off hook input 25:", GPIO.input(25))
						if (timecount >= self.core.inc_timeout):	# timeout
							core.decline_call(call, linphone.Reason.Declined)
							telstat = 5
					except KeyboardInterrupt:	# break with keyboard
						#core.stop_ringing()
						break

				print "\n *** Registered call:", telstat, timecount, call.remote_address.as_string_uri_only(), time.asctime( time.localtime(time.time()) ) 
				# write log-file
				myfile = file('playerphone.log', 'a')   # start log: log-file-handle
				mystring = 'Call registered %s: %s %s - %s \n' % (telstat, timecount, call.remote_address.as_string_uri_only(), time.asctime( time.localtime(time.time()) ) )   # create log
				myfile.write(mystring)   # write log to file
				myfile.close()   # end log: close file
					
				# Off hook  
				if (telstat == 4):
					params = core.create_call_params(call)
					core.accept_call_with_params(call, params)
					telstat = 2
					timestart = time.time()
					print "\a *** incoming call accepted!\n"
					# write log-file
					myfile = file('playerphone.log', 'a')   # start log: log-file-handle
					mystring = 'Call accepted %s: %s - %s \n' % (telstat, call.remote_address.as_string_uri_only(), time.asctime( time.localtime(time.time()) ) )   # create log
					myfile.write(mystring)   # write log to file
					myfile.close()   # end log: close file
				else:
					telstat = 0
					timestart = time.time()
			else:
				core.decline_call(call, linphone.Reason.Declined)
				telstat = 0
				timestart = time.time()
				print "\a *** incoming call declined!\n"
				# write log-file
				myfile = file('playerphone.log', 'a')   # start log: log-file-handle
				mystring = 'Call declined %s: %s - %s \n' % (telstat, call.remote_address.as_string_uri_only(), time.asctime( time.localtime(time.time()) ) )  # create log 
				myfile.write(mystring)   # write log to file
				myfile.close()   # end log: close file
				# send chat message about foreign call
				#chat_room = core.get_chat_room_from_uri(self.whitelist[0])
				#msg = chat_room.create_message(call.remote_address_as_string + ' tried to call')
				#chat_room.send_chat_message(msg)
	# End of call_state_changed


	def handlecall(self, domainStr):
		global c
		c = 0
		last = 1
		countnumber = 0
		number = 0
		global telstat
		global timestart
		timecount = 0
		timestart = time.time()
		timest = 0
		timecnt = 0
		address = domainStr
		sterne = ''
		dialnumber = ''

		print "\n *** Handle:", telstat, address, time.asctime( time.localtime(time.time()) )
		# write log-file
		myfile = file('playerphone.log', 'a')   # start log
		mystring = 'Handle call %s: %s - %s \n' % (telstat, address, time.asctime( time.localtime(time.time()) ) )
		myfile.write(mystring)
		myfile.close()   # end log

		if (telstat == 1):
			if (address != None):
				if (telstat == 1 and GPIO.input(25) == 0):
					self.core.invite(address)
					telstat = 2
					print "**** Handle:", telstat, "start call to", address, time.asctime( time.localtime(time.time()) )
					# write log-file 
					myfile = file('playerphone.log', 'a') 	# start log
					mystring = 'Calling %s: %s - %s \n' % (telstat, address, time.asctime( time.localtime(time.time()) ) )
					myfile.write(mystring)
					myfile.close()	# end log
		if (telstat > 2 or GPIO.input(25) == 1):   # == 3 or hang up
			timestart = time.time()
			self.core.ringer_device = self.core.capture_device
			self.core.terminate_all_calls()	# hang up
			#GPIO.remove_event_detect(23)  # just in case
			print "**** Handle:", telstat, "hang up call to", address, time.asctime( time.localtime(time.time()) )
			# write log-file
			myfile = file('playerphone.log', 'a')	# start log
			mystring = 'Phone: hang up %s: %s - %s \n' % (telstat, address, time.asctime( time.localtime(time.time()) ) )
			myfile.write(mystring)
			myfile.close()	# end log
			telstat = 0    
	# End of handlecall
		

	def configure_sip_account(self, username, password):
		# Configure the SIP account
		proxy_cfg = self.core.create_proxy_config()
		proxy_cfg.identity = 'sip:{username}@fritz.box'.format(username=username)
		proxy_cfg.server_addr = 'sip:fritz.box'
		proxy_cfg.register_enabled = True
		self.core.add_proxy_config(proxy_cfg)
		auth_info = self.core.create_auth_info(username, None, password, None, None, 'fritz.box')
		self.core.add_auth_info(auth_info)
		

	def count(self, pin):
		global timestart
		timestart = time.time()
		global c 
		c = c + 1
		#print "**** c:", c
		
		
	def leiser(self):
		m = alsaaudio.Mixer(control='Headphone')   # defined alsaaudio.Mixer to change volume
		m.setvolume(42,0)	# set volume left
		m.setvolume(42,1)	# set volume right 
		vol = m.getvolume() # get volume float value
		#print "**** volume:", vol
		
	
	def lauter(self):
		m = alsaaudio.Mixer(control='Headphone')   # defined alsaaudio.Mixer to change volume
		m.setvolume(98) # set volume
		
		
	def switchoff(self, pin):
		global player
		global offswitch
		# offswitch = offswitch + 1		# uncomment for shutdown-button use
 		print "\n *** 24: Shutdown-button:", offswitch
		if (offswitch > 4):
			print "**** 24: Would be real shutdown!"
			try:
				player.kill()
			except NameError:
				pass
			self.core.play_local("/home/pi/music/0button2.wav")
			time.sleep(1)
			self.signal_handler(None, None)   # hang up and quit
			GPIO.cleanup()
			os.system("amixer -c 0 sset Headphone 98%,98%")
			os.system("sudo shutdown -hP now")
			sys.exit()
		else:
			print "**** 24: Wait for 5th dial (or fourth press)..."
			self.core.play_local("/home/pi/music/0hup.wav")
			time.sleep(0.2)


	def run(self):
		global player
		global offswitch
		offswitch = 0
		global timestart
		timestart = time.time()
		timecount = 0
		global c
		c = 0
		global telstat
		telstat = 0
		domain = '@fritz.box'
		last = 1
		number = -1
		music = ''  # sound at start: '' for none
		kanal = '-q'
		sterne = ''
		countnumber = 0
		dialnumber = ''
		address = ''
		gpiodetect = 0

		GPIO.add_event_detect(18, GPIO.BOTH, bouncetime=2)	# Rotary (default: without bouncetime)
		GPIO.add_event_detect(25, GPIO.BOTH, bouncetime=800)	# Hook (300)
		GPIO.add_event_detect(24, GPIO.RISING, callback=self.switchoff, bouncetime=900)	# shutdown button (500)

		while not self.quit:
			self.core.iterate()
			time.sleep(0.03)		# 0.03

			try:
				if GPIO.event_detected(18):		# Rotary start
					# offswitch = 0		# uncomment for shutdown-button use INSTEAD of four times 0
					print "\n *** 18: Mainloop Rotary start number:", number, "last:", last, "timecount:", timecount, time.asctime( time.localtime(time.time()) )
					current = GPIO.input(18)
					if(last != current):
						# print "**** Current:", current
						if(current == 0):
							if (gpiodetect == 0):
								gpiodetect = 1
								GPIO.add_event_detect(23, GPIO.BOTH, callback=self.count, bouncetime=5)  # default bouncetime = 5
						else:
							try:
								player.kill()
							except NameError:
								pass
							GPIO.remove_event_detect(23)
							gpiodetect = 0
							number = math.floor(c/2)		# try 2 or 4 or ...
							print "**** countrotary number", number, "(", c, ")", current
							c = 0
						last = GPIO.input(18)
					timestart = time.time()
					
				if (number >= 0):
					# set player			
					if (number > 9):
						number = 0
					if (number == 1):  
						music = "place Radio-URL here"
						offswitch = 0
					elif (number == 2):
						music = "place Radio-URL here"
						offswitch = 0
					elif (number == 3):
						music = "place Radio-URL here"
						offswitch = 0
					elif (number == 4):
						music = "http://bbcmedia.ic.llnwd.net/stream/bbcmedia_radio4fm_mf_p"
						offswitch = 0
					elif (number == 7):		#example: 7 /home/pi/music/7.mp3
						music = "/home/pi/music/" + str(int(number)) + ".mp3"
						offswitch = 0
					elif (number == 0):
						music = "/home/pi/music/0ton.mp3"
						offswitch = offswitch + 1		# four times 0: shutdown
						if (GPIO.input(25) == 1 and offswitch > 2):
							music = ''
							self.switchoff (24)
						else:
							if (kanal == '-q'):
								kanal = '-mq'		# switch to mono
								self.leiser()		
							else: 
								kanal = '-q' 
								self.lauter()		# switch to loud
					else:
						music = ''
						# music = "/home/pi/music/" + str(int(number)) + ".mp3"
						offswitch = 0
					
					# use phone if off hook
					if (GPIO.input(25) == 0):
						music = ''
						offswitch = 0
						print "**** Outgoing number:", number, "c:", c
						countnumber = countnumber+1
						if (countnumber == 1):
							if (number > 0):
								sterne = '**'	# FritzBox intern
								strnumber = str(int(number))
							else:
								sterne = '*31*'	# FritzBox private call '*31*'
								strnumber = ''
						else:
							#if (dialnumber =='' and countnumber == 2 and number > 0):
								#dialnumber = 'LOCALAREACODE'
							strnumber = str(int(number))                    
						dialnumber = dialnumber + strnumber
						print "**** Outgoing:", sterne, "dialnumber:", dialnumber, "strnumber:", strnumber, "number:", number, "countnumber:", countnumber, "25:", GPIO.input(25), "timecount:", timecount, time.asctime( time.localtime(time.time()) )
						number = -1
						timestart = time.time()
						timecount = time.time() - timestart
						
					# start player
					if (music == ''):
						number = -1
					else:
						try:
							player.kill()
						except NameError:
							pass
						player = subprocess.Popen(["mpg123", kanal, music], stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)	
						print ("**** Mainloop play:", number, music, "(",c,")")
						music = ''
						number = -1
						countnumber = 0
						c=0
						GPIO.remove_event_detect(23)  # just in case
						gpiodetect = 0
						#time.sleep(0.3)

				timecount = time.time() - timestart

				# start phonecall
				if (GPIO.input(25) == 0 and countnumber > 0 and timecount > 3):
					print "**** Mainloop nach if 25:", GPIO.input(25), "timecount:", timecount, time.asctime( time.localtime(time.time()) )
					c = 0
					offswitch = 0
					timestart = time.time()
					timecount = time.time() - timestart
					GPIO.remove_event_detect(23)  # just in case
					gpiodetect = 0
					try:
						player.kill()
					except NameError:
						pass
					address = sterne + dialnumber + domain
					print "**** 25: Mainloop phone action timecount:", timecount, "Status:", GPIO.input(25), "0= abgehoben", address, time.asctime( time.localtime(time.time()) )
					if (telstat < 4):
						self.handlecall(address)	# handle the input regarding the call
					sterne = ''
					address = ''
					dialnumber = ''
					countnumber = 0
					number = -1

				if GPIO.event_detected(25):   # Hook switch: 1 = on hook
					offswitch = 0
					timecount = time.time() - timestart
					if (timecount > 1):   # Change to higher number if to much events detected
						GPIO.remove_event_detect(23)  # just in case
						gpiodetect = 0
						telstat = telstat + 1
						print "**** 25: Mainloop phone action Event:", GPIO.input(25), "0=off hook", time.asctime( time.localtime(time.time()) )
						if (telstat == 1 and GPIO.input(25) == 0):
							try:
								player.kill()
							except NameError:
								pass
							self.core.ringer_device = self.core.playback_device
							self.core.play_local("/usr/local/lib/python2.7/dist-packages/linphone/unittests/sounds/ringback.wav")   # line ready
						if (telstat == 1 and GPIO.input(25) == 1):
							telstat = 0
						if (telstat > 0 and telstat < 4 and GPIO.input(25) == 1):
							self.handlecall(domain)	# handle the input regarding the call
						timestart = time.time()
						timecount = time.time() - timestart
						music = ''
						number = -1
						sterne = ''
						address = ''
						dialnumber = ''
						countnumber = 0

			except KeyboardInterrupt:   # keyboard can break
				break

		print ("**** all ENDed")
		try:
			player.kill()
		except NameError:
			pass 
		os.system("amixer -c 0 sset Headphone 98%,98%")
		GPIO.cleanup()

 
def main():
	global telstat
	telstat = 0
	rophopl = RotaryPlayerPhone(username='INTERN NUMBER e.g. 620', password='PASSWORD', whitelist=['sip:**1@fritz.box', 'sip:**2@fritz.box', 'sip:**621@fritz.box'], camera='V4L2: /dev/video0', snd_capture='ALSA: C-Media USB Headphone Set', snd_playback='ALSA: bcm2835 ALSA')
	rophopl.run()
 
main()
